PATH_NOT_FOUND_ERROR = -1
PATH_NOT_FOUND_ERROR_MSG = "Path not found: {} - error: {}"
PATH_LOAD_ERROR = -2
PATH_LOAD_ERROR_MSG = "Failed to load path: {}"
PATH_CREATE_ERROR = -3
PATH_CREATE_ERROR_MSG = "Failed to create path: {} - error: {}"
CONFIGURE_IMAGE_ERROR = -4
CONFIGURE_IMAGE_ERROR_MSG = "Failed to configure image: {} - error: {}"
RESET_IMAGE_ERROR = -5
RESET_IMAGE_ERROR_MSG = "Failed to reset image: {} - error: {}"
